import ctypes
import signal
import sys
import threading


def async_raise(target_tid, exception):
    # Ensuring and releasing GIL are useless since we're not in C
    # gil_state = ctypes.pythonapi.PyGILState_Ensure()
    ret = ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_long(target_tid), ctypes.py_object(exception)
    )
    # ctypes.pythonapi.PyGILState_Release(gil_state)
    if ret == 0:
        raise ValueError("Invalid thread ID %s" % target_tid)
    elif ret > 1:
        ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(target_tid), None)
        raise SystemError("PyThreadState_SetAsyncExc failed")


class SignalTimeout(object):
    def __init__(self, seconds):
        self.seconds = seconds

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.stop)
        if self.seconds > 0:
            signal.setitimer(signal.ITIMER_REAL, self.seconds)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cancel()
        return False

    def cancel(self):
        signal.setitimer(signal.ITIMER_REAL, 0)

    def stop(self, signum, frame):
        raise TimeoutError


class ThreadTimeout(object):
    def __init__(self, seconds):
        self.seconds = seconds
        self.target_tid = threading.current_thread().ident
        self.timer = None

    def __enter__(self):
        self.timer = threading.Timer(self.seconds, self.stop)
        if self.seconds > 0:
            self.timer.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cancel()
        return False

    def cancel(self):
        if self.timer:
            self.timer.cancel()

    def stop(self):
        async_raise(self.target_tid, TimeoutError)


class TimeoutError(Exception):
    def __init__(self, *args, **kwargs):
        default_message = "Timeout Exceeded."
        if not (args or kwargs):
            args = (default_message,)
        super(TimeoutError, self).__init__(*args, **kwargs)


if sys.platform == "win32":
    Timeout = ThreadTimeout
else:
    Timeout = SignalTimeout



import wrapt


def parse_dsn(dsn):
    return dict(attr.split("=") for attr in dsn.split() if "=" in attr)


class CursorProxy(wrapt.ObjectProxy):
    def __init__(self, cursor, connection_proxy):
        super(CursorProxy, self).__init__(cursor)

        self._self_connection = connection_proxy

    @property
    def connection_proxy(self):
        return self._self_connection

    def execute(self, *args, **kwargs):
        self.__wrapped__.execute(*args, **kwargs)


class ConnectionProxy(wrapt.ObjectProxy):
    def __init__(self, connection, args, kwargs):
        super(ConnectionProxy, self).__init__(connection)

        self._self_args = args
        self._self_kwargs = kwargs

    def cursor(self, *args, **kwargs):  # pragma: no cover
        cursor = self.__wrapped__.cursor(*args, **kwargs)
        return CursorProxy(cursor, self)

    @property
    def extract_db(self):
        return self._self_kwargs.get("db", self._self_kwargs.get("database", ""))

    @property
    def extract_hostname(self):
        return self._self_kwargs.get("host", "localhost")

    @property
    def extract_port(self):
        return self._self_kwargs.get("port")


class AdapterProxy(wrapt.ObjectProxy):
    def prepare(self, *args, **kwargs):  # pragma: no cover
        if not args:
            return self.__wrapped__.prepare(*args, **kwargs)

        connection = args[0]

        if isinstance(connection, wrapt.ObjectProxy):
            connection = connection.__wrapped__

        return self.__wrapped__.prepare(connection, *args[1:], **kwargs)



import sys


class StreamToLogger(object):
    def __init__(self, logger):
        self.logger = logger

    def __getattr__(self, name):
        return getattr(sys.__stdout__, name)

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.info(line.rstrip())
        sys.__stdout__.write(buf)



import collections
import sqlparse
import uuid
import wrapt

from .dbapi import AdapterProxy, ConnectionProxy, CursorProxy
from .util import ensure_utf8

Request = collections.namedtuple(
    "Request", ["command", "key", "hostname", "port", "connectionName", "db", "table"]
)


def collect_mysql_metrics(context, trace, instance, args):
    connection = instance.connection_proxy

    db = connection.extract_db
    hostname = connection.extract_hostname
    port = connection.extract_port

    command, table = None, None

    query = sqlparse.parse(args[0])
    if query:
        query = query[0]
        command = query.get_type()
        table = query.get_name()

    request = Request(
        command=ensure_utf8(command),
        key=None,
        hostname=ensure_utf8(hostname),
        port=ensure_utf8(port),
        connectionName=None,
        db=ensure_utf8(db),
        table=ensure_utf8(table),
    )
    request = request._asdict()
    context.iopipe.mark.db_trace(trace, "mysql", request)


def collect_psycopg2_metrics(context, trace, instance, args):
    try:
        from psycopg2.extensions import parse_dsn
    except ImportError:  # pragma: no cover
        from .dbapi import parse_dsn

    connection = instance.connection_proxy
    dsn = parse_dsn(connection.dsn)

    db = dsn.get("dbname")
    hostname = dsn.get("host", "localhost")
    port = dsn.get("port", 5432)

    command, table = None, None

    query = sqlparse.parse(instance.query)

    if not query and args:
        query = sqlparse.parse(args[0])

    if query:
        query = query[0]
        command = query.get_type()
        table = query.get_name()

    request = Request(
        command=ensure_utf8(command),
        key=None,
        hostname=ensure_utf8(hostname),
        port=ensure_utf8(port),
        connectionName=None,
        db=ensure_utf8(db),
        table=ensure_utf8(table),
    )
    request = request._asdict()
    context.iopipe.mark.db_trace(trace, "postgresql", request)


def collect_pymongo_metrics(context, trace, instance, response):
    from pymongo.cursor import Cursor
    from pymongo.results import (
        BulkWriteResult,
        DeleteResult,
        InsertOneResult,
        InsertManyResult,
        UpdateResult,
    )

    command, key = None, None

    for command_name, key_attr, result_type in [
        ("bulk_write", "inserted_count", BulkWriteResult),
        ("delete", "deleted_count", DeleteResult),
        ("insert_one", "inserted_id", InsertOneResult),
        ("insert_many", "inserted_ids", InsertManyResult),
        ("update", ("upserted_id", "modified_count"), UpdateResult),
    ]:
        if isinstance(response, result_type):
            command = command_name
            if isinstance(key_attr, tuple):
                for attr in key_attr:
                    key = getattr(response, attr, None)
                    if key is not None:
                        break
            else:
                key = getattr(response, key_attr, None)
            if key is not None:
                key = str(key)
            break

    if isinstance(response, Cursor):  # pragman: no cover
        command = "find"
        key = response.retrieved

    hostname, port = None, None

    database = getattr(instance, "database", None)
    if database is not None:
        address = getattr(database, "address", None)
        if address is not None:
            hostname, port = address

    request = Request(
        command=ensure_utf8(command),
        key=ensure_utf8(key),
        hostname=ensure_utf8(hostname),
        port=ensure_utf8(port),
        connectionName=None,
        db=ensure_utf8(getattr(database, "name", None)),
        table=ensure_utf8(getattr(instance, "name", None)),
    )
    request = request._asdict()
    context.iopipe.mark.db_trace(trace, "mongodb", request)


def collect_redis_metrics(context, trace, args, connection):
    command, key = None, None
    if len(args) > 0:
        command = args[0]
    if len(args) > 1:
        key = args[1]

    request = Request(
        command=ensure_utf8(command),
        key=ensure_utf8(key),
        hostname=ensure_utf8(connection.get("host", "localhost")),
        port=ensure_utf8(connection.get("port", 6379)),
        connectionName=None,
        db=ensure_utf8(connection.get("db", 0)),
        table=None,
    )
    request = request._asdict()
    context.iopipe.mark.db_trace(trace, "redis", request)


def patch_mysqldb(context):
    """
    Monkey patches mysqldb client, if available. Overloads the
    execute method to add tracing and metrics collection.
    """

    class _CursorProxy(CursorProxy):
        def execute(self, *args, **kwargs):
            if not hasattr(context, "iopipe") or not hasattr(
                context.iopipe, "mark"
            ):  # pragma: no cover
                self.__wrapped__.execute(*args, **kwargs)
                return

            id = ensure_utf8(str(uuid.uuid4()))
            with context.iopipe.mark(id):
                self.__wrapped__.execute(*args, **kwargs)
            trace = context.iopipe.mark.measure(id)
            context.iopipe.mark.delete(id)
            collect_mysql_metrics(context, trace, self, args)

    class _ConnectionProxy(ConnectionProxy):
        def cursor(self, *args, **kwargs):
            cursor = self.__wrapped__.cursor(*args, **kwargs)
            return _CursorProxy(cursor, self)

    def connect_wrapper(wrapped, instance, args, kwargs):
        connection = wrapped(*args, **kwargs)
        return _ConnectionProxy(connection, args, kwargs)

    for module, attr, wrapper in [
        ("MySQLdb", "connect", connect_wrapper),
        ("MySQLdb", "Connection", connect_wrapper),
        ("MySQLdb", "Connect", connect_wrapper),
    ]:
        try:
            wrapt.wrap_function_wrapper(module, attr, wrapper)
        except Exception:  # pragma: no cover
            pass


def patch_psycopg2(context):
    """
    Monkey patches psycopg2 client, if available. Overloads the
    execute method to add tracing and metrics collection.
    """

    class _CursorProxy(CursorProxy):
        def execute(self, *args, **kwargs):
            if not hasattr(context, "iopipe") or not hasattr(
                context.iopipe, "mark"
            ):  # pragma: no cover
                self.__wrapped__.execute(*args, **kwargs)
                return

            id = ensure_utf8(str(uuid.uuid4()))
            with context.iopipe.mark(id):
                self.__wrapped__.execute(*args, **kwargs)
            trace = context.iopipe.mark.measure(id)
            context.iopipe.mark.delete(id)
            collect_psycopg2_metrics(context, trace, self, args)

    class _ConnectionProxy(ConnectionProxy):
        def cursor(self, *args, **kwargs):
            cursor = self.__wrapped__.cursor(*args, **kwargs)
            return _CursorProxy(cursor, self)

    def adapt_wrapper(wrapped, instance, args, kwargs):
        adapter = wrapped(*args, **kwargs)
        return AdapterProxy(adapter) if hasattr(adapter, "prepare") else adapter

    def connect_wrapper(wrapped, instance, args, kwargs):
        connection = wrapped(*args, **kwargs)
        return _ConnectionProxy(connection, args, kwargs)

    def register_type_wrapper(wrapped, instance, args, kwargs):
        def _extract_arguments(obj, scope=None):
            return obj, scope

        obj, scope = _extract_arguments(*args, **kwargs)

        if scope is not None:
            if isinstance(scope, wrapt.ObjectProxy):
                scope = scope.__wrapped__
            return wrapped(obj, scope)

        return wrapped(obj)

    for module, attr, wrapper in [
        ("psycopg2", "connect", connect_wrapper),
        ("psycopg2.extensions", "adapt", adapt_wrapper),
        ("psycopg2.extensions", "register_type", register_type_wrapper),
        ("psycopg2._psycopg", "register_type", register_type_wrapper),
        ("psycopg2._json", "register_type", register_type_wrapper),
    ]:
        try:
            wrapt.wrap_function_wrapper(module, attr, wrapper)
        except Exception:  # pragma: no cover
            pass


def patch_pymongo(context):
    """
    Monkey patches pymongo client, if available. Overloads the
    query methods to add tracing and metrics collection.
    """

    def wrapper(wrapped, instance, args, kwargs):
        if not hasattr(context, "iopipe") or not hasattr(
            context.iopipe, "mark"
        ):  # pragma: no cover
            return wrapped(*args, **kwargs)

        id = ensure_utf8(str(uuid.uuid4()))
        with context.iopipe.mark(id):
            response = wrapped(*args, **kwargs)
        trace = context.iopipe.mark.measure(id)
        context.iopipe.mark.delete(id)
        collect_pymongo_metrics(context, trace, instance, response)
        return response

    for module, attr, _wrapper in [
        ("pymongo.collection", "Collection.find", wrapper),
        ("pymongo.collection", "Collection.bulk_write", wrapper),
        ("pymongo.collection", "Collection.delete_many", wrapper),
        ("pymongo.collection", "Collection.delete_one", wrapper),
        ("pymongo.collection", "Collection.insert_many", wrapper),
        ("pymongo.collection", "Collection.insert_one", wrapper),
        ("pymongo.collection", "Collection.replace_one", wrapper),
        ("pymongo.collection", "Collection.update_many", wrapper),
        ("pymongo.collection", "Collection.update_one", wrapper),
    ]:
        try:
            wrapt.wrap_function_wrapper(module, attr, _wrapper)
        except Exception:  # pragma: no cover
            pass


def patch_pymysql(context):
    """
    Monkey patches pymysql client, if available. Overloads the
    execute method to add tracing and metrics collection.
    """

    class _CursorProxy(CursorProxy):
        def execute(self, *args, **kwargs):
            if not hasattr(context, "iopipe") or not hasattr(
                context.iopipe, "mark"
            ):  # pragma: no cover
                self.__wrapped__.execute(*args, **kwargs)
                return

            id = ensure_utf8(str(uuid.uuid4()))
            with context.iopipe.mark(id):
                self.__wrapped__.execute(*args, **kwargs)
            trace = context.iopipe.mark.measure(id)
            context.iopipe.mark.delete(id)
            collect_mysql_metrics(context, trace, self, args)

    class _ConnectionProxy(ConnectionProxy):
        def cursor(self, *args, **kwargs):
            cursor = self.__wrapped__.cursor(*args, **kwargs)
            return _CursorProxy(cursor, self)

    def connect_wrapper(wrapped, instance, args, kwargs):
        connection = wrapped(*args, **kwargs)
        return _ConnectionProxy(connection, args, kwargs)

    try:
        wrapt.wrap_function_wrapper("pymysql", "connect", connect_wrapper)
    except Exception:  # pragma: no cover
        pass


def patch_redis(context):
    """
    Monkey patches redis client, if available. Overloads the
    execute methods to add tracing and metrics collection.
    """

    def wrapper(wrapped, instance, args, kwargs):
        if not hasattr(context, "iopipe") or not hasattr(
            context.iopipe, "mark"
        ):  # pragma: no cover
            return wrapped(*args, **kwargs)

        id = ensure_utf8(str(uuid.uuid4()))
        with context.iopipe.mark(id):
            response = wrapped(*args, **kwargs)
        trace = context.iopipe.mark.measure(id)
        context.iopipe.mark.delete(id)
        collect_redis_metrics(
            context, trace, args, instance.connection_pool.connection_kwargs
        )
        return response

    def pipeline_wrapper(wrapped, instance, args, kwargs):  # pragma: no cover
        if not hasattr(context, "iopipe") or not hasattr(
            context.iopipe, "mark"
        ):  # pragma: no cover
            return wrapped(*args, **kwargs)

        # We don't need the entire command stack, just collect a stack count
        pipeline_args = ("PIPELINE", ensure_utf8(len(instance.command_stack)))

        id = ensure_utf8(str(uuid.uuid4()))
        with context.iopipe.mark(id):
            response = wrapped(*args, **kwargs)
        trace = context.iopipe.mark.measure(id)
        context.iopipe.mark.delete(id)
        collect_redis_metrics(
            context, trace, pipeline_args, instance.connection_pool.connection_kwargs
        )
        return response

    for module, attr, _wrapper in [
        ("redis.client", "Redis.execute_command", wrapper),
        ("redis.client", "Pipeline.execute", wrapper),
        ("redis.client", "Pipeline.immediate_execute_command", wrapper),
    ]:
        try:
            wrapt.wrap_function_wrapper(module, attr, _wrapper)
        except Exception:  # pragma: no cover
            pass


def restore_mysqldb():
    """Restores mysqldb"""
    try:
        import MySQLdb
    except ImportError:  # pragma: no cover
        pass
    else:
        setattr(
            MySQLdb, "connect", getattr(MySQLdb.connect, "__wrapped__", MySQLdb.connect)
        )


def restore_psycopg2():
    """Restores psycopg2"""
    try:
        import psycopg2
    except ImportError:  # pragma: no cover
        pass
    else:
        setattr(
            psycopg2,
            "connect",
            getattr(psycopg2.connect, "__wrapped__", psycopg2.connect),
        )
        setattr(
            psycopg2.extensions,
            "register_type",
            getattr(
                psycopg2.extensions.register_type,
                "__wrapped__",
                psycopg2.extensions.register_type,
            ),
        )
        setattr(
            psycopg2._psycopg,
            "register_type",
            getattr(
                psycopg2._psycopg.register_type,
                "__wrapped__",
                psycopg2._psycopg.register_type,
            ),
        )
        setattr(
            psycopg2._json,
            "register_type",
            getattr(
                psycopg2._json.register_type,
                "__wrapped__",
                psycopg2._json.register_type,
            ),
        )


def restore_pymongo():
    """Restores pymongo"""
    try:
        from pymongo.collection import Collection
    except ImportError:  # pragma: no cover
        pass
    else:
        for class_method in (
            "bulk_write",
            "delete_many",
            "delete_one",
            "insert_many",
            "insert_one",
            "replace_one",
            "update_many",
            "update_one",
        ):
            if hasattr(getattr(Collection, class_method), "__wrapped__"):
                setattr(
                    Collection,
                    class_method,
                    getattr(Collection, class_method).__wrapped__,
                )


def restore_pymysql():
    """Restores pymysql"""
    try:
        import pymysql
    except ImportError:  # pragma: no cover
        pass
    else:
        setattr(
            pymysql, "connect", getattr(pymysql.connect, "__wrapped__", pymysql.connect)
        )


def restore_redis():
    """Restores the redis client"""
    try:
        from redis.client import Pipeline, Redis
    except ImportError:  # pragma: no cover
        pass
    else:
        if hasattr(Redis.execute_command, "__wrapped__"):
            setattr(Redis, "execute_command", Redis.execute_command.__wrapped__)
        for class_method in ["execute", "immediate_execute_command"]:
            if hasattr(getattr(Pipeline, class_method), "__wrapped__"):
                setattr(
                    Pipeline, class_method, getattr(Pipeline, class_method).__wrapped__
                )


def patch_db_requests(context):
    if not hasattr(context, "iopipe"):
        return

    patch_mysqldb(context)
    patch_psycopg2(context)
    patch_pymongo(context)
    patch_pymysql(context)
    patch_redis(context)


def restore_db_requests():
    restore_mysqldb()
    restore_psycopg2()
    restore_pymongo()
    restore_pymysql()
    restore_redis()



import copy
import json
import logging
import os
import platform
import sys
import time
import traceback

from . import constants
from .monotonic import monotonic
from .plugins import get_plugin_meta
from .send_report import send_report

if sys.platform.startswith("linux"):
    from . import system
else:
    from . import mock_system as system

logger = logging.getLogger(__name__)


class Report(object):
    """
    The report of system status
    """

    def __init__(self, client, context):
        """
        Instantiates a new IOpipe report.

        :param client: The IOpipe client.
        :param context: The AWS Lambda context.
        """
        self.start_time = monotonic()
        self.sent = False
        self.stat_start = system.read_pid_stat("self")

        self.client = client
        self.config = client.config
        self.context = context

        self.custom_metrics = []
        self.db_trace_entries = []
        self.http_trace_entries = []
        self.labels = set()
        self.performance_entries = []
        self.plugins = get_plugin_meta(self.config["plugins"])

        self.report = {
            "client_id": self.config["token"],
            "coldstart": constants.COLDSTART,
            "custom_metrics": self.custom_metrics,
            "environment": {
                "agent": {
                    "load_time": constants.MODULE_LOAD_TIME,
                    "runtime": "python",
                    "version": constants.VERSION,
                },
                "runtime": {
                    "name": platform.python_implementation(),
                    "version": platform.python_version(),
                },
                # DEPRECATED: the following key will be removed in favor of
                # the 'runtime' in the future
                "python": {"version": platform.python_version()},
                "host": {},
                "os": {"linux": {}},
            },
            "errors": {},
            "dbTraceEntries": self.db_trace_entries,
            "httpTraceEntries": self.http_trace_entries,
            "installMethod": self.config.get("install_method"),
            "performanceEntries": self.performance_entries,
            "plugins": self.plugins,
            "processId": constants.PROCESS_ID,
            "timestamp": int(time.time() * 1000),
        }

        if constants.COLDSTART is True:
            constants.COLDSTART = False
            self.labels.add("@iopipe/coldstart")

    def extract_context_data(self):
        """
        Returns the contents of a AWS Lambda context.

        :returns: A dict of relevant context data.
        :rtype: dict
        """
        data = {}
        for k, v in {
            # camel case names in the report to align with AWS standards
            "functionName": "function_name",
            "functionVersion": "function_version",
            "memoryLimitInMB": "memory_limit_in_mb",
            "invokedFunctionArn": "invoked_function_arn",
            "awsRequestId": "aws_request_id",
            "logGroupName": "log_group_name",
            "logStreamName": "log_stream_name",
        }.items():
            if hasattr(self.context, v):
                data[k] = getattr(self.context, v)
        if (
            hasattr(self.context, "invoked_function_arn")
            and "AWS_SAM_LOCAL" in os.environ
        ):
            data["invokedFunctionArn"] = (
                "arn:aws:lambda:local:0:function:%s"
                % data.get("functionName", "unknown")
            )
        if hasattr(self.context, "get_remaining_time_in_millis") and callable(
            self.context.get_remaining_time_in_millis
        ):
            data[
                "getRemainingTimeInMillis"
            ] = self.context.get_remaining_time_in_millis()
        data["traceId"] = os.getenv("_X_AMZN_TRACE_ID", "")
        return data

    def retain_error(self, error, frame=None):
        """
        Adds details of an error to the report.

        :param error: The error exception to add to the report.
        """
        if frame is None:
            stack = traceback.format_exc()
            self.labels.add("@iopipe/error")
        else:
            stack = "\n".join(traceback.format_stack(frame))
            self.labels.add("@iopipe/timeout")
        details = {
            "name": type(error).__name__,
            "message": "{}".format(error),
            "stack": stack,
        }
        self.report["errors"] = details

    def prepare(self, error=None, frame=None):
        """
        Prepare the report to be sent to IOpipe.

        :param error: An optional error to add to report.
        :param frame: A stack frame to add to report in the event of a timeout.
        """
        if error:
            self.retain_error(error, frame)

        self.report["environment"]["host"]["boot_id"] = system.read_bootid()

        # convert labels to list for sending
        self.report["labels"] = list(self.labels)

        meminfo = system.read_meminfo()

        self.report.update(
            {
                "aws": self.extract_context_data(),
                "timestampEnd": int(time.time() * 1000),
            }
        )

        self.report["environment"]["os"].update(
            {
                "cpus": system.read_stat(),
                "freemem": meminfo["MemFree"],
                "hostname": system.read_hostname(),
                "totalmem": meminfo["MemTotal"],
                "usedmem": meminfo["MemTotal"] - meminfo["MemFree"],
            }
        )

        self.report["environment"]["os"]["linux"]["pid"] = {
            "self": {
                "stat": system.read_pid_stat("self"),
                "stat_start": self.stat_start,
                "status": system.read_pid_status("self"),
            }
        }

        self.report["disk"] = system.read_disk()

        self.report["duration"] = int((monotonic() - self.start_time) * 1e9)

    def send(self):
        """
        Sends the report to IOpipe.
        """
        if self.sent is True:
            return
        self.sent = True

        logger.debug("Sending report to IOpipe:")
        logger.debug(json.dumps(self.report, indent=2, sort_keys=True))

        self.client.submit_future(send_report, copy.deepcopy(self.report), self.config)



from distutils.util import strtobool
import os

from iopipe.plugins import Plugin

from .event_types import metrics_for_event_type, metrics_for_response_type


class EventInfoPlugin(Plugin):
    name = "event-info"
    version = "1.3.0"
    homepage = "https://github.com/iopipe/iopipe-python#event-info-plugin"

    def __init__(self, enabled=True):
        """
        Instantiates the event info plugin

        :param enabled: Whether or not event info should be enabled for all
                        invocations. Alternatively this plugin can be enabled/disabled
                        via the `IOPIPE_EVENT_INFO_ENABLED` environment variable.
        :type enabled: bool
        """
        self._enabled = enabled

    @property
    def enabled(self):
        return self._enabled is True or bool(
            strtobool(os.getenv("IOPIPE_EVENT_INFO_ENABLED", "false"))
        )

    def pre_setup(self, iopipe):
        pass

    def post_setup(self, iopipe):
        pass

    def pre_invoke(self, event, context):
        self.context = context

    def post_invoke(self, event, context):
        if self.enabled:
            self.event_type = metrics_for_event_type(event, context)

    def post_response(self, response):
        if self.enabled:
            metrics_for_response_type(self.event_type, self.context, response)
            self.context.iopipe.inject_step_meta(response)

    def pre_report(self, report):
        pass

    def post_report(self, report):
        pass



import io
import sys

PY3 = sys.version_info[0] == 3
PY3 = sys.version_info[0] == 3
PY37 = sys.version_info[0] == 3 and sys.version_info[1] == 7


if PY3:  # pragma: no cover
    from urllib.parse import urlparse

    binary_types = bytes
    string_types = (str,)
    StringIO = io.StringIO
else:  # pragma: no cover
    from urlparse import urlparse  # noqa

    binary_types = str
    string_types = (basestring,)  # noqa
    StringIO = io.BytesIO



from concurrent.futures import Future, wait
from distutils.util import strtobool
import logging
import os

try:
    import cProfile as profile
except ImportError:
    import profile
import tempfile

from iopipe.plugins import Plugin
from iopipe.signer import get_signed_request

from .request import upload_profiler_report

logger = logging.getLogger(__name__)


class ProfilerPlugin(Plugin):
    name = "profiler"
    version = "1.0.0"
    homepage = "https://github.com/iopipe/iopipe-python#profiler-plugin"

    def __init__(self, enabled=False):
        """
        Instantiates the profiler plugin

        :param enabled: Whether or not the profiler should be enabled for all
                        invocations. Alternatively this plugin can be enabled/disabled
                        via the `IOPIPE_PROFILER_ENABLED` environment variable.
        :type enabled: bool
        """
        self._enabled = enabled

    @property
    def enabled(self):
        return self._enabled is True or bool(
            strtobool(os.getenv("IOPIPE_PROFILER_ENABLED", "false"))
        )

    def pre_setup(self, iopipe):
        self.iopipe = iopipe

    def post_setup(self, iopipe):
        pass

    def pre_invoke(self, event, context):
        self.context = context
        self.profile = None
        self.signed_request = None
        self.stats_file = None

        if self.enabled:
            self.signed_request = self.iopipe.submit_future(
                get_signed_request, self.iopipe.config, self.context, ".cprofile"
            )
            self.profile = profile.Profile()
            self.profile.enable()

    def post_invoke(self, event, context):
        if self.profile is not None:
            self.profile.disable()
            self.context.iopipe.label("@iopipe/plugin-profiler")

    def post_response(self, response):
        pass

    def pre_report(self, report):
        if self.profile is not None:
            if self.signed_request is not None:
                if isinstance(self.signed_request, Future):
                    wait([self.signed_request])
                    self.signed_request = self.signed_request.result()
            if (
                self.signed_request is not None
                and "signedRequest" in self.signed_request
            ):
                with tempfile.NamedTemporaryFile(delete=False) as stats_file:
                    self.profile.dump_stats(stats_file.name)
                    self.iopipe.submit_future(
                        upload_profiler_report,
                        self.signed_request["signedRequest"],
                        stats_file.name,
                        self.iopipe.config,
                    )
                    self.stats_file = stats_file.name
                if "jwtAccess" in self.signed_request:
                    plugin = next((p for p in report.plugins if p["name"] == self.name))
                    if "uploads" not in plugin:
                        plugin["uploads"] = []
                    plugin["uploads"].append(self.signed_request["jwtAccess"])

    def post_report(self, report):
        pass



import functools


class Marker(object):
    def __init__(self, timeline, context):
        self.timeline = timeline
        self.context = context
        self.contexts = []

    def __call__(self, name):
        self.contexts.append(name)
        return self

    def __enter__(self):
        if self.contexts:
            self.start(self.contexts[-1])
        return self

    def __exit__(self, type, value, traceback):
        if self.contexts:
            self.end(self.contexts[-1])
            self.contexts.pop()

    def decorator(self, name):
        return MarkerDecorator(self, name)

    def start(self, name):
        self.timeline.mark("start:%s" % name)
        self.context.iopipe.label("@iopipe/plugin-trace")

    def end(self, name):
        self.timeline.mark("end:%s" % name)

    def measure(self, name, start=None, end=None):
        return self.timeline.measure(
            "measure:%s" % name,
            "start:%s" % (start or name),
            "end:%s" % (end or start or name),
        )

    def delete(self, name):
        self.timeline.delete(name)

    def db_trace(self, trace, db_type, request):
        if self.context.instance.report is None:
            return

        entry = trace._asdict()
        entry["type"] = entry.pop("entryType")
        entry["dbType"] = db_type

        if request is not None:
            entry["request"] = request

        self.context.instance.report.db_trace_entries.append(entry)

    def http_trace(self, trace, request, response):
        if self.context.instance.report is None:
            return

        entry = trace._asdict()
        entry["type"] = entry.pop("entryType")

        if request is not None:
            entry["request"] = request

        if response is not None:
            entry["response"] = response

        self.context.instance.report.http_trace_entries.append(entry)


class MarkerDecorator(object):
    def __init__(self, marker, name):
        self.marker = marker
        self.name = name

    def __call__(self, func):
        self.marker.contexts.append(self.name)

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            self.marker.start(self.marker.contexts[-1])
            return_value = func(*args, **kwargs)
            self.marker.end(self.marker.contexts[-1])
            self.marker.contexts.pop()
            return return_value

        return wrapped



import os
import sys

from setuptools import find_packages, setup

try:
    from pypandoc import convert

    README = convert("README.md", "rst")
except (ImportError, OSError):
    README = open(os.path.join(os.path.dirname(__file__), "README.md"), "r").read()

install_requires = ["sqlparse", "wrapt"]
if sys.version_info[0] == 2:
    install_requires.append("futures")

tests_require = [
    "botocore==1.12.162",
    "fakeredis==1.0.3",
    "jmespath>=0.7.1,<1.0.0",
    "mock",
    "mongomock==3.17.0",
    "more-itertools<6.0.0",
    "mysqlclient==1.4.4",
    "psycopg2-binary==2.8.3",
    "pymongo==3.8.0",
    "PyMySQL==0.9.3",
    "pytest==4.1.0",
    "pytest-benchmark==3.2.0",
    "redis==3.3.4",
    "requests",
]

# Lambda doesn't come with sqlite3 support, which is a dependency of pytest-cov, so we
# have to define these dependencies outside the test suite.
coverage_requires = tests_require + ["coverage==5.0a2", "pytest-cov==2.6.1"]

setup(
    name="iopipe",
    version="1.10.2",
    description="IOpipe agent for serverless Application Performance Monitoring",
    long_description=README,
    author="IOpipe",
    author_email="dev@iopipe.com",
    url="https://github.com/iopipe/iopipe-python",
    packages=find_packages(exclude=("tests", "tests.*")),
    extras_require={
        "coverage": coverage_requires,
        "dev": tests_require + ["black==19.3b0", "pre-commit"],
        "local": install_requires
        + ["botocore==1.12.162", "jmespath>=0.7.1,<1.0.0", "requests"],
    },
    install_requires=install_requires,
    setup_requires=["pytest-runner==4.2"],
    tests_require=tests_require,
    zip_safe=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)



from __future__ import division

import os
import socket

MB_FACTOR = float(1 << 20)


def read_bootid():
    """
    Returns the system boot id.

    :returns: The system bood it.
    :rtype: str
    """
    with open("/proc/sys/kernel/random/boot_id", "rb") as bootid_file:
        return bootid_file.readline().decode("ascii")


def read_disk():
    """
    Returns disk usage for /tmp

    :returns: Disk usage (total, used, percentage used)
    :rtype dict
    """
    s = os.statvfs("/tmp")
    return {
        # This should report as 500MB, if not may need to be hardcoded
        # https://aws.amazon.com/lambda/faqs/
        "totalMiB": (s.f_blocks * s.f_bsize) / MB_FACTOR,
        "usedMiB": ((s.f_blocks - s.f_bfree) * s.f_frsize) / MB_FACTOR,
        "usedPercentage": round(
            (((s.f_blocks - s.f_bfree) * s.f_frsize) / (s.f_blocks * s.f_bsize)) * 100,
            2,
        ),
    }


def read_hostname():
    """
    Returns the system hostname.

    :returns: The system hostname.
    :rtype: str
    """
    return socket.gethostname()


def read_meminfo():
    """
    Returns system memory usage information.

    :returns: The system memory usage.
    :rtype: dict
    """
    data = {}
    with open("/proc/meminfo", "rb") as meminfo_file:
        for row in meminfo_file:
            fields = row.split()
            # Example content:
            # MemTotal:                3801016 kB
            # MemFree:                 1840972 kB
            # MemAvailable:        3287752 kB
            # HugePages_Total:             0
            data[fields[0].decode("ascii")[:-1]] = int(fields[1]) * 1024
    return data


def read_pid_stat(pid="self"):
    """
    Returns system process stat information.

    :param pid: The process ID.
    :returns: The system stat information.
    :rtype: dict
    """
    with open("/proc/%s/stat" % (pid,), "rb") as f:
        stat = f.readline().split()
    return {
        "utime": int(stat[13]),
        "stime": int(stat[14]),
        "cutime": int(stat[15]),
        "cstime": int(stat[16]),
    }


def read_pid_status(pid="self"):
    """
    Returns the system process sstatus.

    :param pid: The process ID.
    :returns: The system process status.
    :rtype: dict
    """
    data = {}
    with open("/proc/%s/status" % (pid,), "rb") as status_file:
        for row in status_file:
            fields = row.split()
            if fields and fields[0] in [b"VmRSS:", b"Threads:", b"FDSize:"]:
                try:
                    data[fields[0].decode("ascii")[:-1]] = int(fields[1])
                except ValueError:
                    data[fields[0].decode("ascii")[:-1]] = fields[1].decode("ascii")
    return data


def read_stat():
    """
    Returns the system stat information.

    :returns: The system stat information.
    :rtype: list
    """
    data = []
    with open("/proc/stat", "rb") as stat_file:
        for line in stat_file:
            cpu_stat = line.split()
            if cpu_stat[0][:3] != b"cpu":
                break
            # First cpu line is aggregation of following lines, skip it
            if len(cpu_stat[0]) == 3:
                continue
            data.append(
                {
                    "times": {
                        "user": int(cpu_stat[1]),
                        "nice": int(cpu_stat[2]),
                        "sys": int(cpu_stat[3]),
                        "idle": int(cpu_stat[4]),
                        "irq": int(cpu_stat[6]),
                    }
                }
            )
    return data
